@@ -1,13 +1,27 @@
-from operator import methodcaller
-import time
-
-from ..exceptions import TransportError
-from ..compat import map, string_types, Queue
-
-from .errors import ScanError, BulkIndexError
+#  Licensed to Elasticsearch B.V. under one or more contributor
+#  license agreements. See the NOTICE file distributed with
+#  this work for additional information regarding copyright
+#  ownership. Elasticsearch B.V. licenses this file to you under
+#  the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+# 	http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing,
+#  software distributed under the License is distributed on an
+#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#  KIND, either express or implied.  See the License for the
+#  specific language governing permissions and limitations
+#  under the License.
 
 import logging
+import time
+from operator import methodcaller
 
+from ..compat import Mapping, Queue, map, string_types
+from ..exceptions import NotFoundError, TransportError
+from .errors import BulkIndexError, ScanError
 
 logger = logging.getLogger("elasticsearch.helpers")
 
@@ -26,9 +40,22 @@ def expand_action(data):
     data = data.copy()
     op_type = data.pop("_op_type", "index")
     action = {op_type: {}}
+
+    # If '_source' is a dict use it for source
+    # otherwise if op_type == 'update' then
+    # '_source' should be in the metadata.
+    if (
+        op_type == "update"
+        and "_source" in data
+        and not isinstance(data["_source"], Mapping)
+    ):
+        action[op_type]["_source"] = data.pop("_source")
+
     for key in (
         "_id",
         "_index",
+        "_if_seq_no",
+        "_if_primary_term",
         "_parent",
         "_percolate",
         "_retry_on_conflict",
@@ -37,6 +64,8 @@ def expand_action(data):
         "_type",
         "_version",
         "_version_type",
+        "if_seq_no",
+        "if_primary_term",
         "parent",
         "pipeline",
         "retry_on_conflict",
@@ -45,7 +74,15 @@ def expand_action(data):
         "version_type",
     ):
         if key in data:
-            if key in ["_parent", "_retry_on_conflict", "_routing", "_version", "_version_type"]:
+            if key in {
+                "_if_seq_no",
+                "_if_primary_term",
+                "_parent",
+                "_retry_on_conflict",
+                "_routing",
+                "_version",
+                "_version_type",
+            }:
                 action[op_type][key[1:]] = data.pop(key)
             else:
                 action[op_type][key] = data.pop(key)
@@ -57,97 +94,85 @@ def expand_action(data):
     return action, data.get("_source", data)
 
 
+class _ActionChunker:
+    def __init__(self, chunk_size, max_chunk_bytes, serializer):
+        self.chunk_size = chunk_size
+        self.max_chunk_bytes = max_chunk_bytes
+        self.serializer = serializer
+
+        self.size = 0
+        self.action_count = 0
+        self.bulk_actions = []
+        self.bulk_data = []
+
+    def feed(self, action, data):
+        ret = None
+        raw_data, raw_action = data, action
+        action = self.serializer.dumps(action)
+        # +1 to account for the trailing new line character
+        cur_size = len(action.encode("utf-8")) + 1
+
+        if data is not None:
+            data = self.serializer.dumps(data)
+            cur_size += len(data.encode("utf-8")) + 1
+
+        # full chunk, send it and start a new one
+        if self.bulk_actions and (
+            self.size + cur_size > self.max_chunk_bytes
+            or self.action_count == self.chunk_size
+        ):
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+            self.size, self.action_count = 0, 0
+
+        self.bulk_actions.append(action)
+        if data is not None:
+            self.bulk_actions.append(data)
+            self.bulk_data.append((raw_action, raw_data))
+        else:
+            self.bulk_data.append((raw_action,))
+
+        self.size += cur_size
+        self.action_count += 1
+        return ret
+
+    def flush(self):
+        ret = None
+        if self.bulk_actions:
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+        return ret
+
+
 def _chunk_actions(actions, chunk_size, max_chunk_bytes, serializer):
     """
     Split actions into chunks by number or size, serialize them into strings in
     the process.
     """
-    bulk_actions, bulk_data = [], []
-    size, action_count = 0, 0
+    chunker = _ActionChunker(
+        chunk_size=chunk_size, max_chunk_bytes=max_chunk_bytes, serializer=serializer
+    )
     for action, data in actions:
-        raw_data, raw_action = data, action
-        action = serializer.dumps(action)
-        # +1 to account for the trailing new line character
-        cur_size = len(action.encode("utf-8")) + 1
-
-        if data is not None:
-            data = serializer.dumps(data)
-            cur_size += len(data.encode("utf-8")) + 1
-
-        # full chunk, send it and start a new one
-        if bulk_actions and (
-            size + cur_size > max_chunk_bytes or action_count == chunk_size
-        ):
-            yield bulk_data, bulk_actions
-            bulk_actions, bulk_data = [], []
-            size, action_count = 0, 0
-
-        bulk_actions.append(action)
-        if data is not None:
-            bulk_actions.append(data)
-            bulk_data.append((raw_action, raw_data))
-        else:
-            bulk_data.append((raw_action,))
-
-        size += cur_size
-        action_count += 1
-
-    if bulk_actions:
-        yield bulk_data, bulk_actions
+        ret = chunker.feed(action, data)
+        if ret:
+            yield ret
+    ret = chunker.flush()
+    if ret:
+        yield ret
 
 
-def _process_bulk_chunk(
-    client,
-    bulk_actions,
-    bulk_data,
-    raise_on_exception=True,
-    raise_on_error=True,
-    *args,
-    **kwargs
-):
-    """
-    Send a bulk request to elasticsearch and process the output.
-    """
+def _process_bulk_chunk_success(resp, bulk_data, ignore_status, raise_on_error=True):
     # if raise on error is set, we need to collect errors per chunk before raising them
     errors = []
 
-    try:
-        # send the actual request
-        resp = client.bulk("\n".join(bulk_actions) + "\n", *args, **kwargs)
-    except TransportError as e:
-        # default behavior - just propagate exception
-        if raise_on_exception:
-            raise e
-
-        # if we are not propagating, mark all actions in current chunk as failed
-        err_message = str(e)
-        exc_errors = []
-
-        for data in bulk_data:
-            # collect all the information about failed actions
-            op_type, action = data[0].copy().popitem()
-            info = {"error": err_message, "status": e.status_code, "exception": e}
-            if op_type != "delete":
-                info["data"] = data[1]
-            info.update(action)
-            exc_errors.append({op_type: info})
-
-        # emulate standard behavior for failed actions
-        if raise_on_error:
-            raise BulkIndexError(
-                "%i document(s) failed to index." % len(exc_errors), exc_errors
-            )
-        else:
-            for err in exc_errors:
-                yield False, err
-            return
-
     # go through request-response pairs and detect failures
     for data, (op_type, item) in zip(
         bulk_data, map(methodcaller("popitem"), resp["items"])
     ):
-        ok = 200 <= item.get("status", 500) < 300
-        if not ok and raise_on_error:
+        status_code = item.get("status", 500)
+
+        ok = 200 <= status_code < 300
+        if not ok and raise_on_error and status_code not in ignore_status:
             # include original document source
             if len(data) > 1:
                 item["data"] = data[1]
@@ -162,6 +187,83 @@ def _process_bulk_chunk(
         raise BulkIndexError("%i document(s) failed to index." % len(errors), errors)
 
 
+def _process_bulk_chunk_error(
+    error, bulk_data, ignore_status, raise_on_exception=True, raise_on_error=True
+):
+    # default behavior - just propagate exception
+    if raise_on_exception and error.status_code not in ignore_status:
+        raise error
+
+    # if we are not propagating, mark all actions in current chunk as failed
+    err_message = str(error)
+    exc_errors = []
+
+    for data in bulk_data:
+        # collect all the information about failed actions
+        op_type, action = data[0].copy().popitem()
+        info = {"error": err_message, "status": error.status_code, "exception": error}
+        if op_type != "delete":
+            info["data"] = data[1]
+        info.update(action)
+        exc_errors.append({op_type: info})
+
+    # emulate standard behavior for failed actions
+    if raise_on_error and error.status_code not in ignore_status:
+        raise BulkIndexError(
+            "%i document(s) failed to index." % len(exc_errors), exc_errors
+        )
+    else:
+        for err in exc_errors:
+            yield False, err
+
+
+def _process_bulk_chunk(
+    client,
+    bulk_actions,
+    bulk_data,
+    raise_on_exception=True,
+    raise_on_error=True,
+    ignore_status=(),
+    *args,
+    **kwargs
+):
+    """
+    Send a bulk request to elasticsearch and process the output.
+    """
+    kwargs = _add_helper_meta_to_kwargs(kwargs, "bp")
+
+    if not isinstance(ignore_status, (list, tuple)):
+        ignore_status = (ignore_status,)
+
+    try:
+        # send the actual request
+        resp = client.bulk(*args, body="\n".join(bulk_actions) + "\n", **kwargs)
+    except TransportError as e:
+        gen = _process_bulk_chunk_error(
+            error=e,
+            bulk_data=bulk_data,
+            ignore_status=ignore_status,
+            raise_on_exception=raise_on_exception,
+            raise_on_error=raise_on_error,
+        )
+    else:
+        gen = _process_bulk_chunk_success(
+            resp=resp,
+            bulk_data=bulk_data,
+            ignore_status=ignore_status,
+            raise_on_error=raise_on_error,
+        )
+    for item in gen:
+        yield item
+
+
+def _add_helper_meta_to_kwargs(kwargs, helper_meta):
+    params = (kwargs or {}).pop("params", {})
+    params["__elastic_client_meta"] = (("h", helper_meta),)
+    kwargs["params"] = params
+    return kwargs
+
+
 def streaming_bulk(
     client,
     actions,
@@ -174,6 +276,7 @@ def streaming_bulk(
     initial_backoff=2,
     max_backoff=600,
     yield_ok=True,
+    ignore_status=(),
     *args,
     **kwargs
 ):
@@ -209,6 +312,7 @@ def streaming_bulk(
         2**retry_number``
     :arg max_backoff: maximum number of seconds a retry will wait
     :arg yield_ok: if set to False will skip successful documents in the output
+    :arg ignore_status: list of HTTP status code that you want to ignore
     """
     actions = map(expand_action_callback, actions)
 
@@ -230,6 +334,7 @@ def streaming_bulk(
                         bulk_data,
                         raise_on_exception,
                         raise_on_error,
+                        ignore_status,
                         *args,
                         **kwargs
                     ),
@@ -266,7 +371,7 @@ def streaming_bulk(
                 bulk_actions, bulk_data = to_retry, to_retry_data
 
 
-def bulk(client, actions, stats_only=False, *args, **kwargs):
+def bulk(client, actions, stats_only=False, ignore_status=(), *args, **kwargs):
     """
     Helper for the :meth:`~elasticsearch.Elasticsearch.bulk` api that provides
     a more human friendly interface - it consumes an iterator of actions and
@@ -288,6 +393,7 @@ def bulk(client, actions, stats_only=False, *args, **kwargs):
     :arg actions: iterator containing the actions
     :arg stats_only: if `True` only report number of successful/failed
         operations instead of just number of successful and a list of error responses
+    :arg ignore_status: list of HTTP status code that you want to ignore
 
     Any additional keyword arguments will be passed to
     :func:`~elasticsearch.helpers.streaming_bulk` which is used to execute
@@ -301,7 +407,9 @@ def bulk(client, actions, stats_only=False, *args, **kwargs):
 
     # make streaming_bulk yield successful results so we can count them
     kwargs["yield_ok"] = True
-    for ok, item in streaming_bulk(client, actions, *args, **kwargs):
+    for ok, item in streaming_bulk(
+        client, actions, ignore_status=ignore_status, *args, **kwargs
+    ):
         # go through request-response pairs and detect failures
         if not ok:
             if not stats_only:
@@ -321,6 +429,7 @@ def parallel_bulk(
     max_chunk_bytes=100 * 1024 * 1024,
     queue_size=4,
     expand_action_callback=expand_action,
+    ignore_status=(),
     *args,
     **kwargs
 ):
@@ -341,6 +450,7 @@ def parallel_bulk(
         (`None` if data line should be omitted).
     :arg queue_size: size of the task queue between the main thread (producing
         chunks to send) and the processing threads.
+    :arg ignore_status: list of HTTP status code that you want to ignore
     """
     # Avoid importing multiprocessing unless parallel_bulk is used
     # to avoid exceptions on restricted environments like App Engine
@@ -350,7 +460,7 @@ def parallel_bulk(
 
     class BlockingPool(ThreadPool):
         def _setup_queues(self):
-            super(BlockingPool, self)._setup_queues()
+            super(BlockingPool, self)._setup_queues()  # type: ignore
             # The queue must be at least the size of the number of threads to
             # prevent hanging when inserting sentinel values during teardown.
             self._inqueue = Queue(max(queue_size, thread_count))
@@ -362,7 +472,12 @@ def parallel_bulk(
         for result in pool.imap(
             lambda bulk_chunk: list(
                 _process_bulk_chunk(
-                    client, bulk_chunk[1], bulk_chunk[0], *args, **kwargs
+                    client,
+                    bulk_chunk[1],
+                    bulk_chunk[0],
+                    ignore_status=ignore_status,
+                    *args,
+                    **kwargs
                 )
             ),
             _chunk_actions(
@@ -428,16 +543,37 @@ def scan(
         )
 
     """
-    scroll_kwargs = scroll_kwargs or {}
+    scroll_kwargs = scroll_kwargs.copy() if scroll_kwargs else {}
+    scroll_kwargs["scroll"] = scroll
+    _add_helper_meta_to_kwargs(scroll_kwargs, "s")
 
     if not preserve_order:
         query = query.copy() if query else {}
         query["sort"] = "_doc"
 
+    # Grab options that should be propagated to every
+    # API call within this helper instead of just 'search()'
+    transport_kwargs = {}
+    for key in ("headers", "api_key", "http_auth"):
+        if key in kwargs:
+            transport_kwargs[key] = kwargs[key]
+
+    # If the user is using 'scroll_kwargs' we want
+    # to propagate there too, but to not break backwards
+    # compatibility we'll not override anything already given.
+    if scroll_kwargs is not None and transport_kwargs:
+        for key, val in transport_kwargs.items():
+            scroll_kwargs.setdefault(key, val)
+
     # initial search
-    resp = client.search(
-        body=query, scroll=scroll, size=size, request_timeout=request_timeout, **kwargs
-    )
+    search_kwargs = kwargs.copy()
+    if query:
+        search_kwargs.update(query)
+    search_kwargs["scroll"] = scroll
+    search_kwargs["size"] = size
+    search_kwargs["request_timeout"] = request_timeout
+    _add_helper_meta_to_kwargs(search_kwargs, "s")
+    resp = client.search(**search_kwargs)
     scroll_id = resp.get("_scroll_id")
 
     try:
@@ -445,34 +581,43 @@ def scan(
             for hit in resp["hits"]["hits"]:
                 yield hit
 
+            # Default to 0 if the value isn't included in the response
+            shards_successful = resp["_shards"].get("successful", 0)
+            shards_skipped = resp["_shards"].get("skipped", 0)
+            shards_total = resp["_shards"].get("total", 0)
+
             # check if we have any errors
-            if (resp["_shards"]["successful"] + resp["_shards"]["skipped"]) < resp[
-                "_shards"
-            ]["total"]:
+            if (shards_successful + shards_skipped) < shards_total:
+                shards_message = "Scroll request has only succeeded on %d (+%d skipped) shards out of %d."
                 logger.warning(
-                    "Scroll request has only succeeded on %d (+%d skipped) shards out of %d.",
-                    resp["_shards"]["successful"],
-                    resp["_shards"]["skipped"],
-                    resp["_shards"]["total"],
+                    shards_message,
+                    shards_successful,
+                    shards_skipped,
+                    shards_total,
                 )
                 if raise_on_error:
                     raise ScanError(
                         scroll_id,
-                        "Scroll request has only succeeded on %d (+%d skiped) shards out of %d."
+                        shards_message
                         % (
-                            resp["_shards"]["successful"],
-                            resp["_shards"]["skipped"],
-                            resp["_shards"]["total"],
+                            shards_successful,
+                            shards_skipped,
+                            shards_total,
                         ),
                     )
-            resp = client.scroll(
-                body={"scroll_id": scroll_id, "scroll": scroll}, **scroll_kwargs
-            )
+
+            scroll_kwargs["scroll_id"] = scroll_id
+            resp = client.scroll(**scroll_kwargs)
             scroll_id = resp.get("_scroll_id")
 
     finally:
         if scroll_id and clear_scroll:
-            client.clear_scroll(body={"scroll_id": [scroll_id]}, ignore=(404,))
+            client.clear_scroll(
+                scroll_id=scroll_id,
+                ignore=(404,),
+                params={"__elastic_client_meta": (("h", "s"),)},
+                **transport_kwargs
+            )
 
 
 def reindex(
@@ -483,6 +628,7 @@ def reindex(
     target_client=None,
     chunk_size=500,
     scroll="5m",
+    op_type=None,
     scan_kwargs={},
     bulk_kwargs={},
 ):
@@ -512,6 +658,9 @@ def reindex(
     :arg chunk_size: number of docs in one chunk sent to es (default: 500)
     :arg scroll: Specify how long a consistent view of the index should be
         maintained for scrolled search
+    :arg op_type: Explicit operation type. Defaults to '_index'. Data streams must
+        be set to 'create'. If not specified, will auto-detect if target_index is a
+        data stream.
     :arg scan_kwargs: additional kwargs to be passed to
         :func:`~elasticsearch.helpers.scan`
     :arg bulk_kwargs: additional kwargs to be passed to
@@ -520,18 +669,41 @@ def reindex(
     target_client = client if target_client is None else target_client
     docs = scan(client, query=query, index=source_index, scroll=scroll, **scan_kwargs)
 
-    def _change_doc_index(hits, index):
+    def _change_doc_index(hits, index, op_type):
         for h in hits:
             h["_index"] = index
+            if op_type is not None:
+                h["_op_type"] = op_type
             if "fields" in h:
                 h.update(h.pop("fields"))
             yield h
 
     kwargs = {"stats_only": True}
     kwargs.update(bulk_kwargs)
+
+    is_data_stream = False
+    try:
+        # Verify if the target_index is data stream or index
+        data_streams = target_client.indices.get_data_stream(
+            target_index, expand_wildcards="all"
+        )
+        is_data_stream = any(
+            data_stream["name"] == target_index
+            for data_stream in data_streams["data_streams"]
+        )
+    except (TransportError, KeyError, NotFoundError):
+        # If its not data stream, might be index
+        pass
+
+    if is_data_stream:
+        if op_type not in (None, "create"):
+            raise ValueError("Data streams must have 'op_type' set to 'create'")
+        else:
+            op_type = "create"
+
     return bulk(
         target_client,
-        _change_doc_index(docs, target_index),
+        _change_doc_index(docs, target_index, op_type),
         chunk_size=chunk_size,
         **kwargs
     )
